/* cttest.c - parse /proc/net/ip_conntrack, compute bucket occupation
 *
 * This is how lines look:
 *
 * udp      17 0 src=213.6.73.215 dst=62.104.191.241 sport=1858 dport=53 src=62.104.191.241 dst=213.6.73.215 sport=53 dport=1858 use=1 
 * tcp      6 114 TIME_WAIT src=62.104.211.64 dst=212.7.144.66 sport=34685 dport=80 src=212.7.144.66 dst=62.104.211.64 sport=80 dport=34685 [ASSURED] use=1 
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>

typedef unsigned long long u64;
typedef unsigned long u32;
typedef unsigned short u16;
typedef unsigned char u8;

#ifdef PER_HASH_OUTER
#define PER_HASH_TIMER_1(code) do{code}while(0)
#define PER_HASH_TIMER_2(code) PER_HASH_TIMER(code)
#define PER_HASH_OUTPUT 1
#define PER_HASH_OUTPUT_TYPE "OUTER"
#else
#ifdef PER_HASH_INNER
#define PER_HASH_TIMER_1(code) PER_HASH_TIMER(code)
#define PER_HASH_TIMER_2(code) do{code}while(0)
#define PER_HASH_OUTPUT 1
#define PER_HASH_OUTPUT_TYPE "INNER"
#else
#define PER_HASH_TIMER_1(code) do{total_hash_count++;do{code}while(0);}while (0)
#define PER_HASH_TIMER_2(code) do{code}while(0)
#undef PER_HASH_OUTPUT
#endif
#endif

static u32 total_hash_count;

#ifdef PER_HASH_OUTPUT
static u64 total_hash_time;
#include "rdtsc.c"
#define PER_HASH_TIMER(code) \
	do { \
		u64 timer = rdtsc_read(); \
		do{code}while(0); \
		timer = rdtsc_read() - timer; \
		total_hash_time += timer; \
		total_hash_count++; \
	} while (0)
#endif

/*
 * The key structure is arranged to resemble the kernel tuple layout
 * as of July 2002.
 */
struct ct_key {
	u32 sip;
	u16 sport;
	u32 dip;
	u16 dport;
	u16 proto;
};

static u32 hash_seed;

/*
 * hash_conntrack(), later called "original", is the kernel code
 * as of July 2002.
 */
static u32 hash_conntrack(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ntohl(key->sip + key->dip
	+ key->sport + key->dport
	+ key->proto)
	+ ntohs(key->sport);
);
	return res;
}

/*
 * hash_abcd() has been proposed by Don Cohen
 * http://lists.samba.org/pipermail/netfilter-devel/2002-July/005008.html
 */
static u32 hash_abcd(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = 0x47441DFB * key->sip
	+ 0x57655A7D * key->dip
	+ 0x1C7F1D2D * key->sport
	+ 0xDF91D738 * key->dport
	+ key->proto;
);
	return res;
}

/*
 * hash_abcdef() has later been proposed by Don Cohen; here it is
 * from code by Joakim.
 */
static u32 hash_abcdef(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ( 0x47441DFB * key->sip
	+ 0x57655A7D * key->dip
	+ 0x1C7F1D2D * key->sport
	+ 0xDF91D737 * key->dport
	+ 0x2F6FE8DB * key->proto ) ^ 0x39E5C2D7;
);
	return res;
}

/*
 * hash_abcd_xor(), misnamed abcd_power(), is by Joakim Axelsson and
 * Martin Josefsson, from a series of list mails starting here:
 * http://lists.samba.org/pipermail/netfilter-devel/2002-July/005068.html
 * I took all these functions from their working copy. Thanks.
 */
static u32 hash_abcd_xor(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = (0x47441DFB * key->sip)
	^ (0x57655A7D * key->dip)
	^ (0x1C7F1D2D * key->sport)
	^ (0xDF91D738 * key->dport)
	^ key->proto;
);
	return res;
}

/*
 * hash_rt() was derived from the IP routing code by Joakim and Martin.
 */
static u32 hash_rt(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->dip & 0xF0F0F0F0) >> 4) | ((key->dip & 0x0F0F0F0F) << 4);
	res ^= key->sip ^ key->proto;
	res ^= key->dport ^ key->sport;
	res ^= (res >> 24);
	res ^= (res >> 8);
);
	return res;
}

/*
 * hash_rt_ab() is an obfuscation of hash_rt(). The motto of this
 * function is "the compiler will certainly undo any constant damage."
 * (sorry, could not resists :)
 */
static u32 hash_rt_ab(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->dip & 0xF0F0F0F0) >> 4) | ((key->dip & 0x0F0F0F0F) << 4);
	res ^= key->sip ^ key->proto;
	res ^= key->dport ^ key->sport;
	res ^= 0x47441DFB;
	res ^= (res >> 24);
	res ^= (res >> 8);
	res ^= 0x57655A7D;
);
	return res;
}

/*
 * hash_rt_ab2() is a variant of hash_rt_ab().
 * http://lists.samba.org/pipermail/netfilter-devel/2002-July/005075.html
 */
static u32 hash_rt_ab2(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->sip & 0xF0F0F0F0) >> 4) | ((key->sip & 0x0F0F0F0F) << 4);
	res *= 0x47441DFB;
	res ^= key->dip ^ key->proto;
	res ^= ((key->dport << 16) | key->sport);
	res ^= 0x57655A7D;
	res ^= (res >> 24);
	res ^= (res >> 8);
);
	return res;
}

/*
 * hash_rt_new() is another variant of hash_rt_ab().
 */
static u32 hash_rt_new(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->sip & 0xF0F0F0F0) >> 4) | ((key->sip & 0x0F0F0F0F) << 4);
	res *= 0x47441DFB;
	res ^= key->dip ^ key->proto;
	res ^= (key->dport << 16) ^ key->sport;
	res *= 0x57655A7D;
	res ^= (res >> 24);
	res ^= (res >> 8);
);
	return res;
}

/*
 * hash_crc32() uses the zlib crc32() function over the keys.
 */

#include "crc32.c"

static u32 hash_crc32(struct ct_key *key)
{
	u32 crc;
	u32 v32;
	u16 v16;
PER_HASH_TIMER_1(
	crc = crc32(hash_seed, (const char *) &key->proto, sizeof(key->proto));
	v32 = ntohl(key->sip);
	crc = crc32(crc, (const char *) &v32, sizeof(v32));
	v32 = ntohl(key->dip);
	crc = crc32(crc, (const char *) &v32, sizeof(v32));
	v16 = ntohs(key->sport);
	crc = crc32(crc, (const char *) &v16, sizeof(v16));
	v16 = ntohs(key->dport);
	crc = crc32(crc, (const char *) &v16, sizeof(v16));
);
	return crc;
}

static u32 hash_crc32_2x6(struct ct_key *key)
{
	u32 crc;
PER_HASH_TIMER_1(
	crc = crc32_2x6(hash_seed, (const char *) &key->sip);
);
	return crc;
}

static u32 hash_crc32_bof(struct ct_key *key)
{
	u32 crc;
PER_HASH_TIMER_1(
	crc = crc32_2_6(hash_seed ^ key->sip, (const char *) &key->sport);
);
	return crc;
}

/*
 * hash_random() simulates the optimum, by randomly choosing buckets.
 */
static u32 hash_random(struct ct_key *key)
{
	u32 res;
	PER_HASH_TIMER_1(res = (u32) random(););
	return res;
}

struct hash_def {
	char *name;
	u32 (*hash_f)(struct ct_key *);
	u32 seed;
};

#define NR_HASHES (sizeof(hashes)/sizeof(hashes[0]))

static struct hash_def hashes[] = {
	{ "original",	hash_conntrack, 0 },
	{ "abcd",	hash_abcd,	0 },
	{ "abcdef",	hash_abcdef,	0 },
	{ "abcd_xor",	hash_abcd_xor,	0 },
	{ "rt",		hash_rt,	0 },
	{ "rt_ab",	hash_rt_ab,	0 },
	{ "rt_ab2",	hash_rt_ab2,	0 },
	{ "rt_new",	hash_rt_new,	0 },
	{ "crc32",	hash_crc32,	0 },
	{ "crc32_2x6",	hash_crc32_2x6,	0x1C7F1D2D },
	{ "crc32_bof",	hash_crc32_bof,	0x1C7F1D2D },
	{ "random",	hash_random,	0 },
};

static u32 htable_size = 7168;
static u32 *occupation;
static struct hash_def *test_hash = hashes;

static void do_init(void)
{
	u32 idx;
	occupation = malloc(sizeof(*occupation) * htable_size);
	for (idx=0; idx<htable_size; idx++)
		occupation[idx] = 0;
}

static int display_target = -1;

static int do_count(struct ct_key *key)
{
	u32 idx;
	
	PER_HASH_TIMER_2(idx = test_hash->hash_f(key) % htable_size;);
	occupation[idx]++;
	return ((int)idx) == display_target;
}

static u32 parse_ip(char *t)
{
	char *p = strchr(t, '=');
	if (!p++) return 0;
	return inet_addr(p);
}

static u16 parse_port(char *t)
{
	char *p = strchr(t, '=');
	if (!p++) return 0;
	return htons((u16) strtoul(p, 0, 10));
}

static u32 ignore_tcp_timeout_bound = 0;

static int ignore_by_timeout(struct ct_key *key, char *tok)
{
	return	key->proto == 6
		&& strtoul(tok, 0, 10) < ignore_tcp_timeout_bound;
}

static int unidir;

static int process(char *p)
{
	struct ct_key key[1];
	char *tok;
	int res;

	strtok(p, " \t\n");					/* prot text */
	tok = strtok(0, " \t\n");
	key->proto = (u16) strtoul(tok, 0, 10);			/* prot num */
	tok = strtok(0, " \t\n");				/* timeout */
	if (ignore_by_timeout(key, tok)) return 0;
	switch (key->proto) {
		case 6:  strtok(0, " \t\n");			/* tcp state */
		case 17: break;
		default: return 0;
	}
	/* forward */
	tok = strtok(0, " \t\n"); key->sip = parse_ip(tok);	/* src ip */
	tok = strtok(0, " \t\n"); key->dip = parse_ip(tok);	/* dst ip */
	tok = strtok(0, " \t\n"); key->sport = parse_port(tok);	/* src port */
	tok = strtok(0, " \t\n"); key->dport = parse_port(tok);	/* dst port */
	res = do_count(key);
	if (unidir) return res;
	/* reverse */
	tok = strtok(0, " \t\n");
	if (*tok == '[') tok = strtok(0, " \t\n");
	key->sip = parse_ip(tok);				/* src ip */
	tok = strtok(0, " \t\n"); key->dip = parse_ip(tok);	/* dst ip */
	tok = strtok(0, " \t\n"); key->sport = parse_port(tok);	/* src port */
	tok = strtok(0, " \t\n"); key->dport = parse_port(tok);	/* dst port */
	return res | do_count(key);
}

static int occupation_target = -1;

static void show_occupation(void)
{
	u32 idx, maxlen, use, *agg;
	u64 partial;

	for (idx=0, maxlen=0, use=0; idx<htable_size; idx++) {
		if (occupation[idx] > 0) use++;
		if (occupation[idx] > maxlen)
			maxlen = occupation[idx];
	}
	printf("total buckets:       %lu\n", htable_size);
	printf("total tuples:        %lu (%.02f%%)\n", total_hash_count, (100.0*total_hash_count)/htable_size);
	printf("used buckets:        %lu (%.02f%%)\n", use, (100.0*use)/htable_size);
	printf("maximum list length: %lu\n", maxlen);
	printf("average list length: %.02f\n", (1.0 * total_hash_count) / use);
#ifdef PER_HASH_OUTPUT
	printf("hash cost per tuple: %Lu cyc, %.0f ns, %.0f Mhz on %.0f Mhz CPU (%s)\n",
		total_hash_time / total_hash_count,
		(ns_per_clock * total_hash_time) / total_hash_count,
		(1e3 * total_hash_count) / (ns_per_clock * total_hash_time),
		1e3 / ns_per_clock,
		PER_HASH_OUTPUT_TYPE);
#endif
	agg = malloc(sizeof(*agg) * (maxlen+1));
	for (idx=0; idx<=maxlen; idx++)
		agg[idx] = 0;
	for (idx=0; idx<htable_size; idx++) {
		agg[occupation[idx]]++;
		if (((int)occupation[idx]) == occupation_target) {
			printf("occupation target %d at %lu\n",
				occupation_target, idx);
			occupation_target = -1;
		}
	}
	printf("Count\tLength\tCT\n");
	for (partial=0ULL, idx=maxlen; idx>0; idx--) {
		double d;
		if (agg[idx] == 0) continue;
		partial += idx * agg[idx];
		d = (100.0 * partial) / (1.0 * total_hash_count);
		printf("%lu\t%lu\t%.02f%%\n", agg[idx], idx, d);
	}
	printf("%lu\t0\t100.00%%\n", agg[idx]);
}

static struct hash_def *find_hash(char *name)
{
	u32 idx;
	char *num = 0;
	char *p = name + strlen(name) - 1;

	if ((p-1) >= name && p[-1] == '+' && (*p == 'r' || *p == 'R')) {
		p[-1] = '\0';
		srandom(time(0));
		hash_seed = random();
	} else {
		while (p > name && *p >= '0' && *p <= '9')
			num = p--;
		if (num && p >= name && *p == '+') {
			hash_seed = strtoul(num, 0, 10);
			*p = '\0';
		}
	}
	for (idx=0; idx<NR_HASHES; idx++) {
		if (0 == strcmp(name, hashes[idx].name)) {
			if (!hash_seed)
				hash_seed = hashes[idx].seed;
			return hashes+idx;
		}
	}
	fprintf(stderr, "no hash function named '%s'\n", name);
	return 0;
}

static void usage(void)
{
	fprintf(stderr, "Usage: not written\n");
	exit(1);
}

#include "prime.c"

int main(int argc, char **argv)
{
	int c;
	char buf[512], copy[512];
	char *filename = "/proc/net/ip_conntrack";
	FILE *fp;
	int select_prime = 0;

	while (EOF!=(c=getopt(argc, argv, "f:s:h:i:t:b:c:CPu"))) switch(c) {
		case 'f': filename = optarg; break;
		case 'u': unidir = 1; break;
		case 's': htable_size = strtoul(optarg, 0, 10); break;
		case 'P': select_prime = 1; break;
		case 'h': test_hash = find_hash(optarg); break;
		case 'i': display_target = strtol(optarg, 0, 10); break;
		case 't': occupation_target = strtol(optarg, 0, 10); break;
		case 'b': ignore_tcp_timeout_bound = strtoul(optarg, 0, 10); break;
		case 'c': hash_seed = strtoul(optarg, 0, 10); break;
		case 'C': srandom(time(0)); hash_seed = random(); break;
		default: usage();
	}
	if (!test_hash) return 1;
	fp = (0 == strcmp(filename, "-")) ? stdin : fopen(filename, "r");
	if (!fp) {
		fprintf(stderr, "%s: %s\n", filename, strerror(errno));
		exit(1);
	}
#ifdef PER_HASH_OUTPUT
	rdtsc_cal(100000000);
#endif
	if (select_prime) htable_size = make_prime(htable_size);
	printf("filename: %s\n", filename);
	printf("hash function: %s\n", test_hash->name);
	if (ignore_tcp_timeout_bound)
		printf("ignore TCP conntracks with timeout less than %lu\n",
			ignore_tcp_timeout_bound);
	if (hash_seed)
		printf("hash seed 0x%0lx\n", hash_seed);
	do_init();
	while (fgets(buf, sizeof(buf), fp)) {
		if (display_target != -1)
			strcpy(copy, buf);
		if (process(buf))
			printf("[%d]\t%s", display_target, copy);
	}
	if (fp != stdin) fclose(fp);
	show_occupation();
	return 0;
}
